from __future__ import unicode_literals

import logging
import re

from mopidy import backend, models
from mopidy.models import SearchResult

from mopidy_beets.browsers.albums import (
    AlbumsByArtistBrowser, AlbumsByGenreBrowser, AlbumsByYearBrowser)
from mopidy_beets.translator import assemble_uri, parse_uri


logger = logging.getLogger(__name__)

# match dates of the following format:
#   YYYY, YYYY-MM, YYYY-MM-DD, YYYY/MM, YYYY/MM/DD
DATE_REGEX = re.compile(
    r'^(?P<year>\d{4})(?:[-/](?P<month>\d{1,2})(?:[-/](?P<day>\d{1,2}))?)?$')


class BeetsLibraryProvider(backend.LibraryProvider):

    root_directory = models.Ref.directory(uri='beets:library',
                                          name='Beets library')
    root_categorie_list = [
        ('albums-by-artist', 'Albums by Artist', AlbumsByArtistBrowser),
        ('albums-by-genre', 'Albums by Genre', AlbumsByGenreBrowser),
        ('albums-by-year', 'Albums by Year', AlbumsByYearBrowser),
    ]

    def __init__(self, *args, **kwargs):
        super(BeetsLibraryProvider, self).__init__(*args, **kwargs)
        self.remote = self.backend.beets_api
        self.category_browsers = []
        for key, label, browser_class in self.root_categorie_list:
            ref = models.Ref.directory(name=label, uri=assemble_uri(
                self.root_directory.uri, key))
            browser = browser_class(ref, self.remote)
            self.category_browsers.append(browser)

    def browse(self, uri):
        logger.debug('Browsing Beets at: %s', uri)
        path, item_id = parse_uri(uri, uri_prefix=self.root_directory.uri)
        if path is None:
            logger.error('Beets - failed to parse uri: %s', uri)
            return []
        elif uri == self.root_directory.uri:
            # top level - show the categories
            refs = [browser.ref for browser in self.category_browsers]
            refs.sort(key=lambda item: item.name)
            return refs
        elif path == 'album':
            # show an album
            try:
                album_id = int(item_id)
            except ValueError:
                logger.error('Beets - invalid album ID in URI: %s', uri)
                return []
            tracks = self.remote.get_tracks_by([('album_id', album_id)], True,
                                               ['track+'])
            return [models.Ref.track(uri=track.uri, name=track.name)
                    for track in tracks]
        else:
            # show a generic category directory
            for browser in self.category_browsers:
                if path == parse_uri(browser.ref.uri,
                                     uri_prefix=self.root_directory.uri)[0]:
                    if item_id is None:
                        return browser.get_toplevel()
                    else:
                        return browser.get_directory(item_id)
            else:
                logger.error('Invalid browse URI: %s / %s', uri, path)
                return []

    def search(self, query=None, uris=None, exact=False):
        # TODO: restrict the result to 'uris'
        logger.debug('Beets Query (exact=%s) within "%s": %s',
                     exact, uris, query)
        if not self.remote.has_connection:
            return SearchResult(uri='beets:search-disconnected', tracks=[])

        self._validate_query(query)
        search_list = []
        for (field, values) in query.items():
            for val in values:
                # missing / unsupported fields: uri, performer
                if field == 'any':
                    search_list.append(val)
                elif field == 'album':
                    search_list.append(('album', val))
                elif field == 'artist':
                    search_list.append(('artist', val))
                elif field == 'albumartist':
                    search_list.append(('albumartist', val))
                elif field == 'track_name':
                    search_list.append(('title', val))
                elif field == 'track_no':
                    search_list.append(('track', val))
                elif field == 'composer':
                    search_list.append(('composer', val))
                elif field == 'genre':
                    search_list.append(('genre', val))
                elif field == 'comment':
                    search_list.append(('comments', val))
                elif field == 'date':
                    # supported date formats: YYYY, YYYY-MM, YYYY-MM-DD
                    # Days and months may consist of one or two digits.
                    # A slash (instead of a dash) is acceptable as a separator.
                    match = DATE_REGEX.search(val)
                    if match:
                        # remove None values
                        for key, value in match.groupdict().items():
                            if value:
                                search_list.append((key, int(value)))
                    else:
                        logger.info(
                            'Beets search: ignoring unknown date format (%s). '
                            'It should be "YYYY", "YYYY-MM" or "YYYY-MM-DD".',
                            val)
                else:
                    logger.info('Beets: ignoring unknown query key: %s', field)
                    break
        logger.debug('Search query: %s', search_list)
        tracks = self.remote.get_tracks_by(search_list, exact, [])
        uri = '-'.join([item if isinstance(item, str) else '='.join(item)
                        for item in search_list])
        return SearchResult(uri='beets:search-' + uri, tracks=tracks)

    def lookup(self, uri=None, uris=None):
        logger.debug('Beets lookup: %s', uri or uris)
        if uri:
            # the older method (mopidy < 1.0): return a list of tracks
            # handle one or more tracks given with multiple semicolons
            logger.debug('Beets lookup: %s', uri)
            path, item_id = parse_uri(uri, uri_prefix=self.root_directory.uri)
            if path == 'track':
                tracks = [self.remote.get_track(item_id)]
            elif path == 'album':
                tracks = self.remote.get_tracks_by([('album_id', item_id)],
                                                   True, ('disc+', 'track+'))
            elif path == 'artist':
                artist_tracks = self.remote.get_tracks_by(
                    [('artist', item_id)], True, [])
                composer_tracks = self.remote.get_tracks_by(
                    [('composer', item_id)], True, [])
                # Append composer tracks to the artist tracks (unique items).
                tracks = list(set(artist_tracks + composer_tracks))
                tracks.sort(key=lambda t: (t.date, t.disc_no, t.track_no))
            else:
                logger.info('Unknown Beets lookup URI: %s', uri)
                tracks = []
            # remove occourences of None
            return [track for track in tracks if track]
        else:
            # the newer method (mopidy>=1.0): return a dict of uris and tracks
            return {uri: self.lookup(uri=uri) for uri in uris}

    def get_distinct(self, field, query=None):
        logger.debug('Beets distinct query: %s (uri=%s)', field, query)
        if not self.remote.has_connection:
            return []
        else:
            return self.remote.get_sorted_unique_track_attributes(field)

    def _validate_query(self, query):
        for values in query.values():
            if not values:
                raise LookupError('Missing query')
            for value in values:
                if not value:
                    raise LookupError('Missing query')



from __future__ import unicode_literals

import logging

from mopidy import backend

import pykka

from .client import BeetsRemoteClient
from .library import BeetsLibraryProvider


logger = logging.getLogger(__name__)


class BeetsBackend(pykka.ThreadingActor, backend.Backend):
    uri_schemes = ['beets']

    def __init__(self, config, audio):
        super(BeetsBackend, self).__init__()

        beets_endpoint = 'http://%s:%s' % (
            config['beets']['hostname'], config['beets']['port'])

        self.beets_api = BeetsRemoteClient(beets_endpoint, config['proxy'])
        self.library = BeetsLibraryProvider(backend=self)
        self.playback = BeetsPlaybackProvider(audio=audio, backend=self)
        self.playlists = None


class BeetsPlaybackProvider(backend.PlaybackProvider):

    def translate_uri(self, uri):
        track_id = uri.split(';')[1]
        logger.debug('Getting info for track %s with id %s' % (uri, track_id))
        return self.backend.beets_api.get_track_stream_url(track_id)



import logging
import urllib

from mopidy.models import Album, Artist, Track


logger = logging.getLogger(__name__)


def parse_date(data):
    # use 'original' dates if possible
    if 'original_year' in data:
        day = data.get('original_day', None)
        month = data.get('original_month', None)
        year = data.get('original_year', None)
    elif 'year' in data:
        day = data.get('day', None)
        month = data.get('month', None)
        year = data.get('year', None)
    else:
        return None
    # mopidy accepts dates as 'YYYY' or 'YYYY-MM-DD'
    if day is not None and month is not None:
        return '{year:04d}-{month:02d}-{day:02d}'.format(day=day, month=month,
                                                         year=year)
    else:
        return '{year:04d}'.format(year=year)


def _apply_beets_mapping(target_class, mapping, data):
    """ evaluate a mapping of target keys and their source keys or callables

    'target_class' is the Mopidy model to be used for creating the item.
    'mapping' is a dict of {'target': source}.
    Here 'source' could be one of the following types:
        * string: the key for the corresponding value in 'data'
        * callable: a function with a dict ('data') as its only parameter
    """
    kwargs = {}
    for key, map_value in mapping.items():
        if map_value is None:
            value = None
        elif callable(map_value):
            value = map_value(data)
        else:
            value = data.get(map_value, None)
        # ignore None, empty strings or zeros (e.g. for length)
        if value:
            kwargs[key] = value
    return target_class(**kwargs) if kwargs else None


def _filter_none(values):
    return [value for value in values if value is not None]


def parse_artist(data, name_keyword):
    # see https://docs.mopidy.com/en/latest/api/models/#mopidy.models.Artist
    mapping = {
        'uri': lambda d: assemble_uri('beets:library:artist',
                                      id_value=d[name_keyword]),
        'name': name_keyword,
    }
    if name_keyword == 'artist':
        mapping['sortname'] = 'artist_sort'
        mapping['musicbrainz_id'] = 'mb_artistid'
    elif name_keyword == 'albumartist':
        mapping['sortname'] = 'albumartist_sort'
        mapping['musicbrainz_id'] = 'mb_albumartistid'
    else:
        # others - e.g. composers
        pass
    return _apply_beets_mapping(Artist, mapping, data)


def parse_album(data, api):
    # see https://docs.mopidy.com/en/latest/api/models/#mopidy.models.Album
    # The order of items is based on the above documentation.
    # Attributes without corresponding Beets data are mapped to 'None'.
    mapping = {
        'uri': lambda d: assemble_uri('beets:library:album', id_value=d['id']),
        'name': 'album',
        'artists': lambda d: _filter_none([parse_artist(d, 'albumartist')]),
        'num_tracks': 'tracktotal',
        'num_discs': 'disctotal',
        'date': lambda d: parse_date(d),
        'musicbrainz_id': 'mb_albumid',
        # TODO: 'images' is deprecated since v1.2 - move to Library.get_images
        'images': lambda d, api=api: _filter_none(
            [api.get_album_art_url(d['id'])]),
    }
    return _apply_beets_mapping(Album, mapping, data)


def parse_track(data, api):
    # see https://docs.mopidy.com/en/latest/api/models/#mopidy.models.Track
    # The order of items is based on the above documentation.
    # Attributes without corresponding Beets data are mapped to 'None'.
    mapping = {
        'uri': lambda d: 'beets:library:track;%s' % d['id'],
        'name': 'title',
        'artists': lambda d: _filter_none([parse_artist(d, 'artist')]),
        'album': lambda d, api=api: api.get_album(d['album_id'])
                                    if 'album_id' in d else None,
        'composers': lambda d: _filter_none([parse_artist(d, 'composer')]),
        'performers': None,
        'genre': 'genre',
        'track_no': 'track',
        'disc_no': 'disc',
        'date': lambda d: parse_date(d),
        'length': lambda d: int(d.get('length', 0) * 1000),
        'bitrate': lambda d: int(d.get('bitrate', 0) / 1000),
        'comment': 'comments',
        'musicbrainz_id': 'mb_trackid',
        'last_modified': lambda d: int(d.get('mtime', 0)),
    }
    return _apply_beets_mapping(Track, mapping, data)


def parse_uri(uri, uri_prefix=None):
    """ split a URI into an optional prefix and a value

        The format of a uri is similar to this:
            beets:library:album;Foo%20Bar
        (note the ampersand separating the value from the path)

        uri_prefix (optional):
            * remove the string from the beginning of uri
            * the match is valid only if the prefix is separated from the
              remainder of the URI with a color, an ampersand or it is equal
              to the full URI
            * the function returns 'None' if the uri_prefix cannot be removed
              (you should consider this an error condition)

        The result of the function is a tuple of the uri and the id value.
        In case of an error the result is simply None.
    """
    if ';' in uri:
        result_uri, id_string = uri.split(';', 1)
    else:
        result_uri, id_string = uri, None
    last_path_token = result_uri.split(':')[-1]
    if uri_prefix:
        if uri == uri_prefix:
            result_uri = ''
        elif result_uri.startswith(uri_prefix + ':'):
            result_uri = result_uri[len(uri_prefix) + 1:]
        else:
            # this prefix cannot be splitted
            logger.info('Failed to remove URI prefix (%s): %s',
                        uri_prefix, uri)
            return None, None
    if id_string:
        id_value = urllib.unquote(id_string.encode('ascii')).decode('utf-8')
        # convert track and album IDs to int
        if last_path_token in ('track', 'album'):
            try:
                id_value = int(id_value)
            except ValueError:
                logger.info('Failed to parse integer ID from uri: %s', uri)
                return None, None
    else:
        id_value = None
    return result_uri, id_value


def assemble_uri(*args, **kwargs):
    base_path = ':'.join(args)
    id_value = kwargs.pop('id_value', None)
    if id_value is None:
        return base_path
    else:
        # convert numbers and other non-strings
        if not isinstance(id_value, (str, unicode)):
            id_value = str(id_value)
        id_string = urllib.quote(id_value.encode('utf-8'))
        return '%s;%s' % (base_path, id_string)



from mopidy import models

from mopidy_beets.browsers import GenericBrowserBase
from mopidy_beets.translator import assemble_uri


class AlbumsCategoryBrowser(GenericBrowserBase):

    field = None
    sort_fields = None
    label_fields = None

    def get_toplevel(self):
        keys = self.api.get_sorted_unique_album_attributes(self.field)
        return [models.Ref.directory(name=unicode(key), uri=assemble_uri(
            self.ref.uri, id_value=key)) for key in keys]

    def get_directory(self, key):
        albums = self.api.get_albums_by([(self.field, key)], True,
                                        self.sort_fields)
        return [models.Ref.album(uri=album.uri, name=self._get_label(album))
                for album in albums]


class AlbumsByArtistBrowser(AlbumsCategoryBrowser):
    field = 'albumartist'
    sort_fields = ('original_year+', 'year+', 'album+')

    def _get_label(self, album):
        return album.name


class AlbumsByGenreBrowser(AlbumsCategoryBrowser):
    field = 'genre'
    sort_fields = ('albumartist', 'original_year+', 'year+', 'album+')

    def _get_label(self, album):
        artists = ' / '.join([artist.name for artist in album.artists])
        if artists and album.date:
            return '{0} - {1} ({2})'.format(artists, album.name,
                                            album.date.split('-')[0])
        elif artists:
            return '{0} - {1}'.format(artists, album.name)
        else:
            return album.name


class AlbumsByYearBrowser(AlbumsCategoryBrowser):
    field = 'year'
    sort_fields = ('original_month+', 'original_day+',
                   'month+', 'day+', 'album+')

    def _get_label(self, album):
        artists = ' / '.join([artist.name for artist in album.artists])
        if artists:
            return '{0} - {1}'.format(artists, album.name)
        else:
            return album.name
